#!/usr/bin/sh
# Copyright start
# Copyright (C) 2008 - 2024 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end

# Configuration script for cyops-postgresql

fresh_install(){

    # Start postgres
    systemctl enable postgresql-$ver_new

    # initdb
    if [ ! "$(ls -A /$data_dir_new/)" ]; then
        /usr/pgsql-$ver_new/bin/postgresql-$ver_new-setup initdb
    fi

    # Add our custom configs
    yes | install -o postgres -g postgres /opt/cyops/configs/pgsql/pg_hba.conf $data_dir_new/
    yes | install -o postgres -g postgres /opt/cyops/configs/pgsql/postgresql.conf $data_dir_new/

    # Setup ssl form cyops-certs
    yes | cp /etc/pki/cyops/server.leaf.* /var/lib/pgsql/ssl/
    chmod 600 /var/lib/pgsql/ssl/server.leaf.*
    chown -R postgres: /var/lib/pgsql/ssl/

    systemctl start postgresql-$ver_new
    systemctl status postgresql-$ver_new -l

    sudo -u postgres psql -U postgres -c "DROP USER IF EXISTS cyberpgsql"
    set +x
    echo "sudo -u postgres psql -U postgres -c "CREATE USER cyberpgsql WITH SUPERUSER PASSWORD '*****'""
    sudo -u postgres psql -U postgres -c "CREATE USER cyberpgsql WITH SUPERUSER PASSWORD 'qaplplqa'"
    set -x
}

execute_file()
{
    local file_to_execute=$1
    if [ -f $file_to_execute ];then
        sudo -u postgres ./$file_to_execute
    else
        echo "File [$file_to_execute] not found"
    fi
    return
}

handle_noreplace_files()
{
    file=/opt/cyops/configs/pgsql/postgresql.conf

    # modify file to support upgrade from PG 12 to PG 16

    sudo sed -i "s/^\s*data_directory \+\= \+'\/var\/lib\/pgsql\/$ver_prev\/data' *$/data_directory = '\/var\/lib\/pgsql\/$ver_new\/data'/g"  $file

    sudo sed -i "s/^\s*hba_file \+\= \+'\/var\/lib\/pgsql\/$ver_prev\/data\/pg_hba\.conf' *$/hba_file = '\/var\/lib\/pgsql\/$ver_new\/data\/pg_hba.conf'/g"  $file

    sudo sed -i "s/^\s*ident_file \+\= \+'\/var\/lib\/pgsql\/$ver_prev\/data\/pg_ident\.conf' *$/ident_file = '\/var\/lib\/pgsql\/$ver_new\/data\/pg_ident.conf'/g"  $file
  
    sudo sed -i "s/^\s*external_pid_file \+\= \+'\/var\/run\/$ver_prev-data\.pid' *$/external_pid_file = '\/var\/run\/$ver_new-data.pid'/g"  $file

}

update_memory_files_values(){
    local f_postgres_conf="postgresql.conf"
    local f_postgres_conf_rpmnew="postgresql.conf.rpmnew"
    local d_postgres_conf="/opt/cyops/configs/pgsql/"
    local a_list="shared_buffers temp_buffers max_prepared_transactions work_mem maintenance_work_mem max_stack_depth ssl_cert_file ssl_key_file listen_addresses port"
    local f_cmd_cp="/bin/cp"
    $f_cmd_cp "$d_postgres_conf/$f_postgres_conf" "$d_postgres_conf/$f_postgres_conf.bkp"
    if [ -f "$d_postgres_conf/$f_postgres_conf_rpmnew" ]; then
        for s_var in $a_list;
        do
	         s_to_update_value=$(grep "^$s_var" "$d_postgres_conf/$f_postgres_conf")
	         sed -i "/^$s_var/c\\$s_to_update_value" "$d_postgres_conf/$f_postgres_conf_rpmnew"
        done
        yes | $f_cmd_cp "$d_postgres_conf/$f_postgres_conf_rpmnew" "$d_postgres_conf/$f_postgres_conf"
        echo "Successfully updated the $d_postgres_conf/$f_postgres_conf file"
        return
    fi
    echo "The $d_postgres_conf/$f_postgres_conf_rpmnew file not found."
}
update_pg_hba_conf_for_pg14(){
    #
    # From 7.2.2, postgresql upgrade from 12 to 14 and also password method change from trust to md5.
    # While installing the cyops-auth package it try to connect to database using password if md5 is already in \
    # pg_hba.conf file instead of trust.
    #
    local f_cmd_psql="/usr/bin/psql"
    local f_cmd_sed="/usr/bin/sed"
    local s_trust="trust"
    local s_md5="md5"
    local i_psql_version=$(${f_cmd_psql} --version | awk '{print $3}' | cut -d'.' -f1)
    local f_pg_hba_conf="/var/lib/pgsql/${i_psql_version}/data/pg_hba.conf"
    ${f_cmd_sed} -i "s/${s_trust}/${s_md5}/g" $f_pg_hba_conf
    /usr/bin/systemctl restart postgresql-${i_psql_version}
}

update_pg_repo() {
    pg_old_version_repo="
[fsr-pgdg$ver_prev]
name=FortiSOAR PostgreSQL$ver_prev repository
baseurl=https://\$product_yum_server/\$fsr_release_version/third-party/postgres/pgdg$ver_prev/
enabled=1
gpgcheck=1
gpgkey=https://\$product_yum_server/\$fsr_release_version/third-party/postgres/RPM-GPG-KEY-PGDG-$ver_prev"

    if ! grep -q "fsr-pgdg$ver_prev" /etc/yum.repos.d/fsr-third-party.repo; then
        echo "pg$ver_prev repo not set. Setting"
        echo "$pg_old_version_repo" >>/etc/yum.repos.d/fsr-third-party.repo
    fi

}

install_old_pg() {
    echo "Installing postgresql-$ver_prev"
    update_pg_repo
    yum clean all
    yum install postgresql$ver_prev-server -y -q
}

uninstall_old_pg() {
    echo "Uninstalling old pg"
    yum remove postgresql$ver_prev-server postgresql$ver_prev-libs -y -q
    rm -rf /var/lib/pgsql/$ver_prev
}

remove_old_pg_repo_entry() {
    grep "^\[fsr-pgdg$ver_prev\]" /etc/yum.repos.d/fsr-third-party.repo >/dev/null

    # Deleting lines for old PgSQL
    if [ $? -eq 0 ]; then
        sed -i "/^\[fsr-pgdg$ver_prev\]/,/^gpgkey/d" /etc/yum.repos.d/fsr-third-party.repo
    fi
}

cleanup_old_pg() {
    uninstall_old_pg
    remove_old_pg_repo_entry
}

restore_postgres_hba_files() {
    echo "Restoring pg_hba conf"
    local d_postgres_conf="/var/lib/pgsql/$ver_new/data/"
    local f_pg_hba_conf="pg_hba.conf"

    # This is backup file location
    local d_postgres_conf_bkp="/var/lib/pgsql/$ver_prev/data/"

    # Copied previously pg_hba.conf to new
    yes | cp $d_postgres_conf_bkp/$f_pg_hba_conf $d_postgres_conf
}

restore_postgres_config_params_docker() {
    echo "Restoring configuration parameters from old postgres"
    local d_postgres_conf="/var/lib/pgsql/$ver_new/data/"
    local f_postgres_conf="postgresql.conf"

    local d_postgres_conf_bkp="/var/lib/pgsql/$ver_prev/data/"

    local a_list="shared_buffers temp_buffers max_prepared_transactions work_mem maintenance_work_mem max_stack_depth ssl_cert_file ssl_key_file listen_addresses port"
    local f_cmd_cp="/bin/cp"
    local f_cmd_chown="/usr/bin/chown"

    $f_cmd_cp "$d_postgres_conf/$f_postgres_conf" "$d_postgres_conf/$f_postgres_conf.bkp"
    $f_cmd_chown postgres:postgres "$d_postgres_conf/$f_postgres_conf.bkp"

    if [ -f "$d_postgres_conf_bkp/$f_postgres_conf" ]; then
        for s_var in $a_list; do
            s_to_update_value=$(grep "^$s_var" "$d_postgres_conf_bkp/$f_postgres_conf")
            sed -i "/^$s_var/c\\$s_to_update_value" "$d_postgres_conf/$f_postgres_conf"
        done
    else
        echo "The $d_postgres_conf_bkp/$f_postgres_conf file not found."
    fi
    echo "Configuration parameters from old postgres restored"
}


upgrade(){

    d_old_pgsql_bin="/usr/pgsql-$ver_prev/bin/"
    d_new_pgsql_bin="/usr/pgsql-$ver_new/bin/"

    # Required for docker upgrade pgsql ,Install old pgsql if not present 
    if [ ! -d "$d_old_pgsql_bin" ];then
        install_old_pg
    fi

    # initdb
    if [ ! "$(ls -A $data_dir_new/)" ]; then
        /usr/pgsql-$ver_new/bin/postgresql-$ver_new-setup initdb
    fi

    systemctl stop postgresql-$ver_prev

    # Need to run pg_upgrade as non-root
    # We dont have home dir for postgres user, hence creating it
    home_dir_postgres=/home/postgres
    mkdir $home_dir_postgres
    chown postgres:postgres $home_dir_postgres

    # Required for running pg_upgrade command as postgres user
    # pg_upgrade commmand will connect to the old and new servers several times, 
    # so we need to set authentication to peer in pg_hba.conf 
    # and this is reverted later in this function.
    sed -i --follow-symlinks 's/md5/trust/g' /var/lib/pgsql/$ver_prev/data/pg_hba.conf
    sed -i --follow-symlinks 's/md5/trust/g' /var/lib/pgsql/$ver_new/data/pg_hba.conf

    cd $home_dir_postgres
    sudo -u postgres /usr/pgsql-$ver_new/bin/pg_upgrade \
            --old-bindir=$d_old_pgsql_bin \
            --new-bindir=$d_new_pgsql_bin \
            --old-datadir=/var/lib/pgsql/$ver_prev/data/ \
            --new-datadir=$data_dir_new/

    # pg_hha.conf and postgresql.conf are noreplace files. If there is manual change to these
    # files before upgrade, then we need to handle the upgrade changes for it.

    # Currently, we are handling noreplace files to support upgrade from PG 11 to 12
    handle_noreplace_files

    # Setup ssl form cyops-certs
    yes | cp /etc/pki/cyops/server.leaf.* /var/lib/pgsql/ssl/
    chmod 600 /var/lib/pgsql/ssl/server.leaf.*
    chown -R postgres: /var/lib/pgsql/ssl/

    # replace postgresql.conf with new postgresql.conf: Done
    update_memory_files_values

    # Copy postgresql conf files for docker 
    # ToDo : move postgresql conf files to /opt/cyops
    if [ -f "$f_flag_docker" ];then
        echo "copy the postgresql.conf file to /opt/cyops/configs for docker"
        yes | cp /opt/cyops/postgresql.conf /opt/cyops/configs/pgsql
        yes | cp /opt/cyops/pg_hba.conf /opt/cyops/configs/pgsql
    fi
     
    # Add our custom configs
    yes | install -o postgres -g postgres /opt/cyops/configs/pgsql/pg_hba.conf $data_dir_new/
    yes | install -o postgres -g postgres /opt/cyops/configs/pgsql/postgresql.conf $data_dir_new/

    
    if [ -f "$f_flag_docker" ]; then
         # restore previous version pg_hba file changes to new version if any.
        restore_postgres_hba_files
        # restore previous version postgresql.conf file changes to new version if any.
        restore_postgres_config_params_docker
    fi

    systemctl enable postgresql-$ver_new
    systemctl start postgresql-$ver_new

    # Add Password for postgres user same as system device uuid.
    s_uuid=$(/bin/csadm license --get-device-uuid)
    sudo -Hiu postgres psql -U postgres -c "ALTER USER postgres WITH ENCRYPTED PASSWORD '$s_uuid';"

    # change pg_hba.conf replace trust with md5
    update_pg_hba_conf_for_pg14

    execute_file analyze_new_cluster.sh
    execute_file delete_old_cluster.sh
    cleanup_old_pg
}

check_extension_exists(){
    extension_name=$1
    source /opt/cyops/adm/transaction.sh

    PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d postgres -c "select * from pg_extension where extname = '$extension_name';" | grep "$extension_name"
    return $?
}

install_pg_squeeze_pg_repack(){

    pg_squeeze="pg_squeeze-master"
    pg_squeeze_zip=${pg_squeeze}.zip
    pg_repack="pg_repack-1.5.0"
    pg_repack_zip=${pg_repack}.zip
    pg_squeeze_dir="/opt/cyops/configs/pgsql/pg_squeeze"
    pg_repack_dir="/opt/cyops/configs/pgsql/pg_repack"
    flag_restart_postgresql="false"
    f_wget="/bin/wget"
    # use release directory value from /etc/yum/vars/fsr_release_version to downlaod postgresql extensions
    d_fsr_release_version="$(cat /etc/yum/vars/fsr_release_version)"
    cat /etc/environment
    source /etc/environment
    
    #Need to update PATH to include PGSQL BIN 
    export PATH=$PATH:/usr/pgsql-$ver_new/bin/

    if [ "$flag_ignore_ssl_check_for_repo" = "true" ]; then 
        f_wget="$f_wget --no-check-certificate"
    fi

    check_extension_exists "pg_squeeze"
    if [ $? != 0 ];then
        # Install pg_squeeze
        mkdir -p $pg_squeeze_dir
        $f_wget https://$product_yum_server/$d_fsr_release_version/third-party/postgres/pg_squeeze/$pg_squeeze_zip -P $pg_squeeze_dir

        cd $pg_squeeze_dir
        unzip $pg_squeeze_zip
        cd $pg_squeeze_dir/$pg_squeeze
        make
        make install

        # set the shared_preload_libraries value to pg_squeeze
        sed -i "/shared_preload_libraries/c\shared_preload_libraries = 'pg_squeeze' " /var/lib/pgsql/$ver_new/data/postgresql.conf
        flag_restart_postgresql="true"
    else
        echo "pg_squeeze is already installed."

    fi

    check_extension_exists "pg_repack"
    if [ $? != 0 ];then
        # Install pq_repack
        mkdir -p $pg_repack_dir
        $f_wget https://$product_yum_server/$d_fsr_release_version/third-party/postgres/pg_repack/$pg_repack_zip -P $pg_repack_dir
        cd $pg_repack_dir
        unzip $pg_repack_zip
        cd $pg_repack_dir/$pg_repack
        make 
        make install 
        flag_restart_postgresql="true"
    else
        echo "pg_repack is already installed."    
    fi

    if [ $flag_restart_postgresql = "true" ];then
        systemctl restart postgresql-$ver_new
        systemctl status postgresql-$ver_new -l
    fi
}

check_version(){

    f_flag_docker="/data/.do-not-delete-flag-fsr-in-docker"
    

    for ver in $ver_list; do

        f_old_psql="/usr/pgsql-$ver/bin/psql"

        if [ -f "$f_flag_docker" ];then
            f_old_psql="/var/lib/pgsql/$ver/data/postgresql.conf"
        fi

        if [ -f "$f_old_psql" ];then
            # Old version detected
            # Its an upgrade from old version of cyops
            rm -f $cyops_pg_configured_dir/cyops_pg_${ver}_configured
            ver_prev=$ver
            return 1
        fi
    done

    # Its a fresh install
    return 2
}

# Commented out the call for this method, can be uncommented once we change the postgresql encryption method
update_postgres_encryption() {
    pg_conf=$data_dir_new/postgresql.conf
    pg_hba=$data_dir_new/pg_hba.conf
    grep -e "^password_encryption" /var/lib/pgsql/16/data/postgresql.conf | grep -e "md5"
    if [ $? = 0 ]; then
        echo "Updating the PostgreSQL to start using scram-sha-256 inplace of md5 as encryption method"
        # If postgresql.conf contain above then its an upgrade instance.
        # For Fresh install we do have trust at this level, change is made at config_vm level.
        # For Upgrade it will always be md5 for version 7.5.0 which is a supported path for 7.6.0
        sed -i '/^password_encryption\s*=\s*md5\s*$/s//password_encryption = scram-sha-256/' $pg_conf
        systemctl restart postgresql-$ver_new
        . /opt/cyops/adm/transaction.sh
        PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d venom -c "ALTER USER cyberpgsql WITH ENCRYPTED PASSWORD '$s_pg_pwd';"
        PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d venom -c "ALTER USER postgres WITH ENCRYPTED PASSWORD '$s_pg_pwd';"
        sed -i 's/md5/scram-sha-256/g' $pg_hba
    fi
}

#######
# Main
#######
# Space separated list of versions of pg supported by cyops
ver_list="14"
ver_new=$1
data_dir_new=/var/lib/pgsql/$ver_new/data/
cyops_pg_configured_dir=/opt/cyops/configs
cyops_pg_configured=$cyops_pg_configured_dir/cyops_pg_${ver_new}_configured
utils_restore="/opt/cyops/adm/subcmd/utils_restore"

common(){
	. /opt/cyops/adm/transaction.sh
    PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d postgres -c "CREATE EXTENSION IF NOT EXISTS hstore;"
    PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d postgres -c 'CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'
    PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d postgres -c 'CREATE EXTENSION IF NOT EXISTS "pg_squeeze";'
    PGPASSWORD=${s_pg_pwd} psql -U cyberpgsql -d postgres -c 'CREATE EXTENSION IF NOT EXISTS "pg_repack";'
    touch $cyops_pg_configured
}

if [ -f "$cyops_pg_configured" ];then
    # Same version is installed.
    # Its probably upgrade of same cyops version
    # No op
    echo "Found $cyops_pg_configured"
    echo "Looks like postgresql $ver_new is already configured"
    echo "Skipping the configuration"
    install_pg_squeeze_pg_repack
    common
    # # Adding post common as pgcrypto is required.
    # update_postgres_encryption
    exit 0
fi

# Detect whether it is fresh install or an upgrade
check_version
if [ 2 = $? ];then
    fresh_install
elif [ 1 = $? ];then
    upgrade
fi
install_pg_squeeze_pg_repack
# update_postgres_encryption
common
exit 0
